/*******************************************************************************
 * Copyright (c) 2011-2013 EclipseSource Muenchen GmbH and others.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 * Edgar Mueller - initial API and implementation
 ******************************************************************************/
package org.eclipse.emf.ecp.view.internal.rule;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;

/**
 * <p>
 * Convenience class for managing settings.
 * </p>
 * <p>
 * Provides an equals method that returns {@code true} if the {@link EObject} and the
 * {@link org.eclipse.emf.ecore.EStructuralFeature} are the same.
 * </p>
 * 
 * @author emueller
 */
public class UniqueSetting {

	private final EObject eObject;
	private final EAttribute attribute;

	/**
	 * Convenience method for creating a setting.
	 * 
	 * @param eObject
	 *            an {@link EObject}
	 * @param attribute
	 *            an {@link EAttribute} of the given {@code eObject}
	 * @return the constructed setting
	 */
	public static UniqueSetting createSetting(EObject eObject, EAttribute attribute) {
		return new UniqueSetting(eObject, attribute);
	}

	/**
	 * Constructor.
	 * 
	 * @param eObject
	 *            an {@link EObject}
	 * @param attribute
	 *            an {@link EAttribute} of the given {@code eObject}
	 */
	protected UniqueSetting(EObject eObject, EAttribute attribute) {
		this.eObject = eObject;
		this.attribute = attribute;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (attribute == null ? 0 : attribute.hashCode());
		result = prime * result + (eObject == null ? 0 : eObject.hashCode());
		return result;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final UniqueSetting other = (UniqueSetting) obj;
		if (attribute == null) {
			if (other.attribute != null) {
				return false;
			}
		} else if (!attribute.equals(other.attribute)) {
			return false;
		}
		if (eObject == null) {
			if (other.eObject != null) {
				return false;
			}
		} else if (!eObject.equals(other.eObject)) {
			return false;
		}
		return true;
	}

	/**
	 * Returns the {@link EObject}.
	 * 
	 * @return the {@link EObject}
	 */
	public EObject getEObject() {
		return eObject;
	}

	/**
	 * Returns the {@link EAttribute} of the {@link EObject}.
	 * 
	 * @return the {@link EAttribute}
	 */
	public EAttribute getEAttribute() {
		return attribute;
	}
}
